@@ -13,6 +13,7 @@ package ${package};
 
 use strict;
 use warnings;
+# @@protoc_insertion_point(after_pragmas)
 use MIME::Base64 qw();
 use Google::ProtocolBuffers::Dynamic;
 
@@ -20,9 +21,13 @@ my $gpd = Google::ProtocolBuffers::Dynamic->new;
 
 ${load_blobs}
 
+# @@protoc_insertion_point(after_loading)
+
 $gpd->map(
 ${mappings});
 
+# @@protoc_insertion_point(after_mapping)
+
 undef $gpd;
 
 1;
@@ -54,7 +59,7 @@ sub generate {
             $dump =~ s{$}{,};
             # the '+' is to make sure Perl::Critic does not consider
             # this an anonymous subroutine
-            $dump =~ s[^(\s+)\s({)][\1+\2]mg;
+            $dump =~ s[^(\s+)\s(\{)][$1+$2]mg;
             $dump
         } @{$args{mappings}};
     };
@@ -215,7 +220,7 @@ Google::ProtocolBuffers::Dynamic::MakeModule
 
 =head1 VERSION
 
-version 0.41_01
+version 0.41_02
 
 =head1 AUTHOR
 
